import os
from typing import List, Tuple
from itertools import chain

from llama_parse import LlamaParse

from autorag.data.parse.base import parser_node
from autorag.utils.util import process_batch, get_event_loop


@parser_node
def llama_parse(
	data_path_list: List[str],
	batch: int = 8,
	use_vendor_multimodal_model: bool = False,
	vendor_multimodal_model_name: str = "openai-gpt4o",
	use_own_key: bool = False,
	vendor_multimodal_api_key: str = None,
	**kwargs,
) -> Tuple[List[str], List[str], List[int]]:
	"""
	Parse documents to use llama_parse.
	LLAMA_CLOUD_API_KEY environment variable should be set.
	You can get the key from https://cloud.llamaindex.ai/api-key

	:param data_path_list: The list of data paths to parse.
	:param batch: The batch size for parse documents. Default is 8.
	:param use_vendor_multimodal_model: Whether to use the vendor multimodal model. Default is False.
	:param vendor_multimodal_model_name: The name of the vendor multimodal model. Default is "openai-gpt4o".
	:param use_own_key: Whether to use the own API key. Default is False.
	:param vendor_multimodal_api_key: The API key for the vendor multimodal model.
	:param kwargs: The extra parameters for creating the llama_parse instance.
	:return: tuple of lists containing the parsed texts, path and pages.
	"""
	if use_vendor_multimodal_model:
		kwargs = _add_multimodal_params(
			kwargs,
			use_vendor_multimodal_model,
			vendor_multimodal_model_name,
			use_own_key,
			vendor_multimodal_api_key,
		)

	parse_instance = LlamaParse(**kwargs)

	tasks = [
		llama_parse_pure(data_path, parse_instance) for data_path in data_path_list
	]
	loop = get_event_loop()
	results = loop.run_until_complete(process_batch(tasks, batch))

	del parse_instance

	texts, path, pages = (list(chain.from_iterable(item)) for item in zip(*results))

	return texts, path, pages


async def llama_parse_pure(
	data_path: str, parse_instance
) -> Tuple[List[str], List[str], List[int]]:
	documents = await parse_instance.aload_data(data_path)

	texts = list(map(lambda x: x.text, documents))
	path = [data_path] * len(texts)
	pages = list(range(1, len(documents) + 1))

	return texts, path, pages


def _add_multimodal_params(
	kwargs,
	use_vendor_multimodal_model,
	vendor_multimodal_model_name,
	use_own_key,
	vendor_multimodal_api_key,
) -> dict:
	kwargs["use_vendor_multimodal_model"] = use_vendor_multimodal_model
	kwargs["vendor_multimodal_model_name"] = vendor_multimodal_model_name

	def set_multimodal_api_key(
		multimodal_model_name: str = "openai-gpt4o", _api_key: str = None
	) -> str:
		if multimodal_model_name in ["openai-gpt4o", "openai-gpt-4o-mini"]:
			_api_key = (
				os.getenv("OPENAI_API_KEY", None) if _api_key is None else _api_key
			)
			if _api_key is None:
				raise KeyError(
					"Please set the OPENAI_API_KEY in the environment variable OPENAI_API_KEY "
					"or directly set it on the config YAML file."
				)
		elif multimodal_model_name in ["anthropic-sonnet-3.5"]:
			_api_key = (
				os.getenv("ANTHROPIC_API_KEY", None) if _api_key is None else _api_key
			)
			if _api_key is None:
				raise KeyError(
					"Please set the ANTHROPIC_API_KEY in the environment variable ANTHROPIC_API_KEY "
					"or directly set it on the config YAML file."
				)
		elif multimodal_model_name in ["gemini-1.5-flash", "gemini-1.5-pro"]:
			_api_key = (
				os.getenv("GEMINI_API_KEY", None) if _api_key is None else _api_key
			)
			if _api_key is None:
				raise KeyError(
					"Please set the GEMINI_API_KEY in the environment variable GEMINI_API_KEY "
					"or directly set it on the config YAML file."
				)
		elif multimodal_model_name in ["custom-azure-model"]:
			raise NotImplementedError(
				"Custom Azure multimodal model is not supported yet."
			)
		else:
			raise ValueError("Invalid multimodal model name.")

		return _api_key

	if use_own_key:
		api_key = set_multimodal_api_key(
			vendor_multimodal_model_name, vendor_multimodal_api_key
		)
		kwargs["vendor_multimodal_api_key"] = api_key

	return kwargs



import abc
import logging
from typing import Dict

import pandas as pd
from llama_index.core.llms import LLM

from autorag import generator_models
from autorag.schema import BaseModule
from autorag.utils import result_to_dataframe
from autorag.utils.cast import cast_retrieved_contents

logger = logging.getLogger("AutoRAG")


class BasePassageCompressor(BaseModule, metaclass=abc.ABCMeta):
	def __init__(self, project_dir: str, *args, **kwargs):
		logger.info(
			f"Initialize passage compressor node - {self.__class__.__name__} module..."
		)

	def __del__(self):
		logger.info(
			f"Deleting passage compressor node - {self.__class__.__name__} module..."
		)

	def cast_to_run(self, previous_result: pd.DataFrame, *args, **kwargs):
		logger.info(
			f"Running passage compressor node - {self.__class__.__name__} module..."
		)
		assert (
			"query" in previous_result.columns
		), "previous_result must contain 'query' column."
		assert len(previous_result) > 0, "previous_result must have at least one row."

		queries = previous_result["query"].tolist()

		return queries, cast_retrieved_contents(previous_result)


class LlamaIndexCompressor(BasePassageCompressor, metaclass=abc.ABCMeta):
	param_list = ["prompt", "chat_prompt", "batch"]

	def __init__(self, project_dir: str, **kwargs):
		"""
		Initialize passage compressor module.

		:param project_dir: The project directory
		:param llm: The llm name that will be used to summarize.
			The LlamaIndex LLM model can be used in here.
		:param kwargs: Extra parameter for init llm
		"""
		super().__init__(project_dir)
		kwargs_dict = dict(
			filter(lambda x: x[0] not in self.param_list, kwargs.items())
		)
		llm_name = kwargs_dict.pop("llm")
		self.llm: LLM = make_llm(llm_name, kwargs_dict)

	def __del__(self):
		del self.llm
		super().__del__()

	@result_to_dataframe(["retrieved_contents"])
	def pure(self, previous_result: pd.DataFrame, *args, **kwargs):
		queries, retrieved_contents = self.cast_to_run(previous_result)
		param_dict = dict(filter(lambda x: x[0] in self.param_list, kwargs.items()))
		result = self._pure(queries, retrieved_contents, **param_dict)
		return list(map(lambda x: [x], result))


def make_llm(llm_name: str, kwargs: Dict) -> LLM:
	if llm_name not in generator_models:
		raise KeyError(
			f"{llm_name} is not supported. "
			"You can add it manually by calling autorag.generator_models."
		)
	return generator_models[llm_name](**kwargs)



import importlib
from typing import Callable, Dict


def dynamically_find_function(key: str, target_dict: Dict) -> Callable:
	if key in target_dict:
		module_path, func_name = target_dict[key]
		module = importlib.import_module(module_path)
		func = getattr(module, func_name)
		return func
	else:
		raise KeyError(f"Input module or node {key} is not supported.")


def get_support_modules(module_name: str) -> Callable:
	support_modules = {
		# parse
		"langchain_parse": ("autorag.data.parse", "langchain_parse"),
		"clova": ("autorag.data.parse.clova", "clova_ocr"),
		"llamaparse": ("autorag.data.parse.llamaparse", "llama_parse"),
		"table_hybrid_parse": (
			"autorag.data.parse.table_hybrid_parse",
			"table_hybrid_parse",
		),
		# chunk
		"llama_index_chunk": ("autorag.data.chunk", "llama_index_chunk"),
		"langchain_chunk": ("autorag.data.chunk", "langchain_chunk"),
		# query_expansion
		"query_decompose": ("autorag.nodes.queryexpansion", "QueryDecompose"),
		"hyde": ("autorag.nodes.queryexpansion", "HyDE"),
		"pass_query_expansion": (
			"autorag.nodes.queryexpansion",
			"PassQueryExpansion",
		),
		"multi_query_expansion": (
			"autorag.nodes.queryexpansion",
			"MultiQueryExpansion",
		),
		"QueryDecompose": ("autorag.nodes.queryexpansion", "QueryDecompose"),
		"HyDE": ("autorag.nodes.queryexpansion", "HyDE"),
		"PassQueryExpansion": (
			"autorag.nodes.queryexpansion",
			"PassQueryExpansion",
		),
		"MultiQueryExpansion": (
			"autorag.nodes.queryexpansion",
			"MultiQueryExpansion",
		),
		# retrieval
		"bm25": ("autorag.nodes.lexicalretrieval", "BM25"),
		"BM25": ("autorag.nodes.lexicalretrieval", "BM25"),
		"vectordb": ("autorag.nodes.semanticretrieval", "VectorDB"),
		"VectorDB": ("autorag.nodes.semanticretrieval", "VectorDB"),
		"hybrid_rrf": ("autorag.nodes.hybridretrieval", "HybridRRF"),
		"HybridRRF": ("autorag.nodes.hybridretrieval", "HybridRRF"),
		"hybrid_cc": ("autorag.nodes.hybridretrieval", "HybridCC"),
		"HybridCC": ("autorag.nodes.hybridretrieval", "HybridCC"),
		# passage_augmenter
		"prev_next_augmenter": (
			"autorag.nodes.passageaugmenter",
			"PrevNextPassageAugmenter",
		),
		"PrevNextPassageAugmenter": (
			"autorag.nodes.passageaugmenter",
			"PrevNextPassageAugmenter",
		),
		"pass_passage_augmenter": (
			"autorag.nodes.passageaugmenter",
			"PassPassageAugmenter",
		),
		"PassPassageAugmenter": (
			"autorag.nodes.passageaugmenter",
			"PassPassageAugmenter",
		),
		# passage_reranker
		"monot5": ("autorag.nodes.passagereranker", "MonoT5"),
		"MonoT5": ("autorag.nodes.passagereranker", "MonoT5"),
		"tart": ("autorag.nodes.passagereranker.tart", "Tart"),
		"Tart": ("autorag.nodes.passagereranker.tart", "Tart"),
		"upr": ("autorag.nodes.passagereranker", "Upr"),
		"Upr": ("autorag.nodes.passagereranker", "Upr"),
		"koreranker": ("autorag.nodes.passagereranker", "KoReranker"),
		"KoReranker": ("autorag.nodes.passagereranker", "KoReranker"),
		"pass_reranker": ("autorag.nodes.passagereranker", "PassReranker"),
		"PassReranker": ("autorag.nodes.passagereranker", "PassReranker"),
		"cohere_reranker": ("autorag.nodes.passagereranker", "CohereReranker"),
		"CohereReranker": ("autorag.nodes.passagereranker", "CohereReranker"),
		"rankgpt": ("autorag.nodes.passagereranker", "RankGPT"),
		"RankGPT": ("autorag.nodes.passagereranker", "RankGPT"),
		"jina_reranker": ("autorag.nodes.passagereranker", "JinaReranker"),
		"JinaReranker": ("autorag.nodes.passagereranker", "JinaReranker"),
		"colbert_reranker": ("autorag.nodes.passagereranker", "ColbertReranker"),
		"ColbertReranker": ("autorag.nodes.passagereranker", "ColbertReranker"),
		"sentence_transformer_reranker": (
			"autorag.nodes.passagereranker",
			"SentenceTransformerReranker",
		),
		"SentenceTransformerReranker": (
			"autorag.nodes.passagereranker",
			"SentenceTransformerReranker",
		),
		"flag_embedding_reranker": (
			"autorag.nodes.passagereranker",
			"FlagEmbeddingReranker",
		),
		"FlagEmbeddingReranker": (
			"autorag.nodes.passagereranker",
			"FlagEmbeddingReranker",
		),
		"flag_embedding_llm_reranker": (
			"autorag.nodes.passagereranker",
			"FlagEmbeddingLLMReranker",
		),
		"FlagEmbeddingLLMReranker": (
			"autorag.nodes.passagereranker",
			"FlagEmbeddingLLMReranker",
		),
		"time_reranker": ("autorag.nodes.passagereranker", "TimeReranker"),
		"TimeReranker": ("autorag.nodes.passagereranker", "TimeReranker"),
		"openvino_reranker": ("autorag.nodes.passagereranker", "OpenVINOReranker"),
		"OpenVINOReranker": ("autorag.nodes.passagereranker", "OpenVINOReranker"),
		"voyageai_reranker": ("autorag.nodes.passagereranker", "VoyageAIReranker"),
		"VoyageAIReranker": ("autorag.nodes.passagereranker", "VoyageAIReranker"),
		"mixedbreadai_reranker": (
			"autorag.nodes.passagereranker",
			"MixedbreadAIReranker",
		),
		"MixedbreadAIReranker": (
			"autorag.nodes.passagereranker",
			"MixedbreadAIReranker",
		),
		"flashrank_reranker": ("autorag.nodes.passagereranker", "FlashRankReranker"),
		"FlashRankReranker": ("autorag.nodes.passagereranker", "FlashRankReranker"),
		# passage_filter
		"pass_passage_filter": ("autorag.nodes.passagefilter", "PassPassageFilter"),
		"similarity_threshold_cutoff": (
			"autorag.nodes.passagefilter",
			"SimilarityThresholdCutoff",
		),
		"similarity_percentile_cutoff": (
			"autorag.nodes.passagefilter",
			"SimilarityPercentileCutoff",
		),
		"recency_filter": ("autorag.nodes.passagefilter", "RecencyFilter"),
		"threshold_cutoff": ("autorag.nodes.passagefilter", "ThresholdCutoff"),
		"percentile_cutoff": ("autorag.nodes.passagefilter", "PercentileCutoff"),
		"PassPassageFilter": ("autorag.nodes.passagefilter", "PassPassageFilter"),
		"SimilarityThresholdCutoff": (
			"autorag.nodes.passagefilter",
			"SimilarityThresholdCutoff",
		),
		"SimilarityPercentileCutoff": (
			"autorag.nodes.passagefilter",
			"SimilarityPercentileCutoff",
		),
		"RecencyFilter": ("autorag.nodes.passagefilter", "RecencyFilter"),
		"ThresholdCutoff": ("autorag.nodes.passagefilter", "ThresholdCutoff"),
		"PercentileCutoff": ("autorag.nodes.passagefilter", "PercentileCutoff"),
		# passage_compressor
		"tree_summarize": ("autorag.nodes.passagecompressor", "TreeSummarize"),
		"pass_compressor": ("autorag.nodes.passagecompressor", "PassCompressor"),
		"refine": ("autorag.nodes.passagecompressor", "Refine"),
		"longllmlingua": ("autorag.nodes.passagecompressor", "LongLLMLingua"),
		"TreeSummarize": ("autorag.nodes.passagecompressor", "TreeSummarize"),
		"Refine": ("autorag.nodes.passagecompressor", "Refine"),
		"LongLLMLingua": ("autorag.nodes.passagecompressor", "LongLLMLingua"),
		"PassCompressor": ("autorag.nodes.passagecompressor", "PassCompressor"),
		# prompt_maker
		"fstring": ("autorag.nodes.promptmaker", "Fstring"),
		"long_context_reorder": ("autorag.nodes.promptmaker", "LongContextReorder"),
		"window_replacement": ("autorag.nodes.promptmaker", "WindowReplacement"),
		"Fstring": ("autorag.nodes.promptmaker", "Fstring"),
		"LongContextReorder": ("autorag.nodes.promptmaker", "LongContextReorder"),
		"WindowReplacement": ("autorag.nodes.promptmaker", "WindowReplacement"),
		"chat_fstring": ("autorag.nodes.promptmaker", "ChatFstring"),
		"ChatFstring": ("autorag.nodes.promptmaker", "ChatFstring"),
		# generator
		"llama_index_llm": ("autorag.nodes.generator", "LlamaIndexLLM"),
		"vllm": ("autorag.nodes.generator", "Vllm"),
		"openai_llm": ("autorag.nodes.generator", "OpenAILLM"),
		"vllm_api": ("autorag.nodes.generator", "VllmAPI"),
		"LlamaIndexLLM": ("autorag.nodes.generator", "LlamaIndexLLM"),
		"Vllm": ("autorag.nodes.generator", "Vllm"),
		"OpenAILLM": ("autorag.nodes.generator", "OpenAILLM"),
		"VllmAPI": ("autorag.nodes.generator", "VllmAPI"),
	}
	return dynamically_find_function(module_name, support_modules)


def get_support_nodes(node_name: str) -> Callable:
	support_nodes = {
		"query_expansion": (
			"autorag.nodes.queryexpansion.run",
			"run_query_expansion_node",
		),
		"semantic_retrieval": (
			"autorag.nodes.semanticretrieval.run",
			"run_semantic_retrieval_node",
		),
		"dense_retrieval": (
			"autorag.nodes.semanticretrieval.run",
			"run_semantic_retrieval_node",
		),
		"lexical_retrieval": (
			"autorag.nodes.lexicalretrieval.run",
			"run_lexical_retrieval_node",
		),
		"sparse_retrieval": (
			"autorag.nodes.lexicalretrieval.run",
			"run_lexical_retrieval_node",
		),
		"hybrid_retrieval": (
			"autorag.nodes.hybridretrieval.run",
			"run_hybrid_retrieval_node",
		),
		"generator": ("autorag.nodes.generator.run", "run_generator_node"),
		"prompt_maker": ("autorag.nodes.promptmaker.run", "run_prompt_maker_node"),
		"passage_filter": (
			"autorag.nodes.passagefilter.run",
			"run_passage_filter_node",
		),
		"passage_compressor": (
			"autorag.nodes.passagecompressor.run",
			"run_passage_compressor_node",
		),
		"passage_reranker": (
			"autorag.nodes.passagereranker.run",
			"run_passage_reranker_node",
		),
		"passage_augmenter": (
			"autorag.nodes.passageaugmenter.run",
			"run_passage_augmenter_node",
		),
	}
	return dynamically_find_function(node_name, support_nodes)
